!   Name: read_srb_rel3_sw_3hrly.f90
!   Purpose:  This code reads the GEWEX Release 3.00 3hrly shortwave fluxes and other variables
!   then converts the variables to true 1x1 degree fields.
!
!   Variables:
!   Name      Type                 Description
!   sw_val_in real                 11 scaled fluxes and extra variables set up as:
!             (44016,8,31,11)        (1) All Sky TOA downward flux (toa_down)
!                                    (2) All Sky TOA upward flux  (toa_up) 
!                                    (3) All Sky Surface downward flux  (sfc_down)
!                                    (4) All Sky Surface upward flux  (sfc_up)
!                                    (5) Clear Sky TOA upward flux  (clr_toa_up)
!                                    (6) Clear Sky Surface downward flux  (clr_sfc_down)
!                                    (7) Clear Sky Surface upward flux   (clr_sfc_up)
!                                    (8) All Sky Global PAR  (par)
!                                    (9) Cloud Fraction  (cld_frac) 
!                                   (10) Cosine Solar Zenith Angle  (cos_sza)
!                                   (11) Average cosine solar zenith angle (ave_cos_sza)
!   sclfac    real(11)              Scaling Factor for fluxes and parameters
!   sw_var    real                  Intermediate array of true fluxes and parameters
!   year      integer               Year of data set
!   month     integer               Month of data set 
!   nvars     integer               Number of variables
!   ngrids    integer               Number of grids
!   misval    real                  The value of undefined variables
!   jdays     integer               The Julian days covered in the month
!   ig        integer               Current region number 
!   var0      real (44016)          Input field on the nested grid 
!   var1      real (64800)          Intermediate 1x1 field, 1-dimensional array
!   var2      real (360,180)        Output field on 1x1 grid, 2-dimensional array
!
!   The layout of the nested grid is described in the readme files.
!   both, input (nested) and output 1x1 grids have the same orientation:
!   start at the greenwich meridian-south pole and go east and north.
!
!   Internal Routines:
!   Subroutine Regrid- Regrids a field from the nested grid to an equal-angle 1x1 grid
!   Subroutine Unnest- Replicates field values for the Regrid routine.
!   Subroutine native_4byte_int - Converts 4-byte integers from big Endian to little Endian
!   Subroutine native_2byte_int - Converts 2-byte integers from big Endian to little Endian
!   Subroutine native_4byte_real - Converts 4-byte reals from big Endian to little Endian
!
!   Input files:
!   srb_rel3_sw_3hrly.nml - namelist file to select year, month, input & output
!                      paths and which variables to extract
!   srb_rel3.0_shortwave_3hrly_YYYYMM.binary -data file where YYYY is 4-digit year and
!                      MM is 2-digit month.  This will be filled by the program.
!
!   Output files (if variable is chosen in namelist):
!     set ascii and/or binary logical variable in namelist file to true for output
!   toa_down_3hrly_YYYYMM.ascii  (binary)
!   toa_up_3hrly_YYYYMM.ascii  (binary)
!   sfc_down_3hrly_YYYYMM.ascii  (binary)
!   sfc_up_3hrly_YYYYMM.ascii  (binary)
!   clr_toa_up_3hrly_YYYYMM.ascii  (binary)
!   clr_sfc_down_3hrly_YYYYMM.ascii  (binary) 
!   clr_sfc_up_3hrly_YYYYMM.ascii  (binary)
!   par_3hrly_YYYYMM.ascii  (binary) 
!   cld_frac_3hrly_YYYYMM.ascii  (binary) 
!   cos_sza_3hrly_YYYYMM.ascii  (binary) 
!   ave_cos_sza_3hrly_YYYYMM.ascii  (binary) 
    
       program read_srb_rel3_sw_3hrly

       implicit none

       integer, parameter :: rkind=selected_real_kind(6), ikind=selected_int_kind(6)
       integer(ikind), parameter :: nreg=44016, sreg=64800, nlon=360, nlat=180, nv=11

       character (len=200) :: infile, outfile, outfile2, path_in, path_out
       character (len=244) :: header
       character (len=13), dimension (nv) :: var_header
       character(len=4) :: cyear
       character(len=2) :: cmon

       integer(ikind) :: nrg, srg, lon, lat, jrec 
       integer(ikind) :: breg, ereg, gmtime(8)
       real (rkind) :: var0(nreg), var1(sreg), var2(nlon,nlat)


       integer(ikind) :: year, month, nvars, numgmt, ndays, ig, ngrids, iflx, reg
       integer(ikind) :: yr, mon, hr, day, i, irec
       real (rkind) :: sw_var(nreg), sclfac(nv)
       integer (kind=2) :: sw_val_in(nreg,8,31,nv)
       real (rkind) :: vernum, vertab, misval
       integer(ikind), dimension (31) :: jdays

       logical :: toa_down, toa_up, sfc_down, sfc_up, clr_toa_up, &
               clr_sfc_down, clr_sfc_up, par, cld_frac, cos_sza, ave_cos_sza
       logical :: var_list(nv), ascii, binary, little_endian

!   In namelist file, srb_rel3_sw_3hrly.nml, set the desired variables to 'true' 
!   to convert to one degree. Set little_endian to 'true' if your machine expects
!   variables to be in that byte order.  Note: If possible for that case, it would 
!   be better to use a compiler option to read as Big Endian than to use the subroutines.
!   Specify output type by setting ascii and/or binary to 'true'.

       namelist/time_vars/yr, mon, path_in, path_out, little_endian, ascii, binary, &
               toa_down, toa_up, sfc_down, sfc_up, clr_toa_up, clr_sfc_down, &
               clr_sfc_up, par, cld_frac, cos_sza, ave_cos_sza

       var_header=(/'toa_down_    ', 'toa_up_      ', 'sfc_down_    ', &
          'sfc_up_      ', 'clr_toa_up_  ', 'clr_sfc_down_', 'clr_sfc_up_  ', &
          'par_         ', 'cld_frac_    ', 'cos_sza_     ', 'ave_cos_sza_ '/)

       print*, '*****************************************************************'
       print*, '*                                                               *'
       print*, '*                                                               *'
       print*, '* Data Set srb_rel3.0_shortwave_3hrly Read Software             *'
       print*, '*                                                               *'
       print*, '* Version: 1.0                                                  *'
       print*, '*                                                               *'
       print*, '* Contact: Atmospheric Science Data Center                     *'
       print*, '* User and Data Services Office                       *'
       print*, '* Mail Stop 157D                                                *'
       print*, '* 2 South Wright Street                                         *'
       print*, '* NASA Langley Research Center                                  *'
       print*, '* Hampton, Virginia 23681-2199                                  *'
       print*, '* U.S.A.                                                        *'
       print*, '*                                                               *'
       print*, '* E-mail: larc@eos.nasa.gov                                     *'
       print*, '* Phone: (757)864-8656                                          *'
       print*, '* FAX: (757)864-8807                                            *'
       print*, '*                                                               *'
       print*, '*****************************************************************'
                          
!   Open and read namelist file
       open (1,file='srb_rel3_sw_3hrly.nml')
       read (1,nml=time_vars)
       close (1)

       write (cyear,"(I4)") yr
       write (cmon,"(I2.2)") mon
       
!   Re-assign logicals from namelist 
       var_list(1)=toa_down
       var_list(2)=toa_up
       var_list(3)=sfc_down
       var_list(4)=sfc_up
       var_list(5)=clr_toa_up
       var_list(6)=clr_sfc_down
       var_list(7)=clr_sfc_up
       var_list(8)=par
       var_list(9)=cld_frac
       var_list(10)=cos_sza
       var_list(11)=ave_cos_sza
          
!   Open input file
       infile=trim(path_in)//'srb_rel3.0_shortwave_3hrly_'//cyear//cmon//'.binary'

!   Read header information
       if (.not. little_endian) then
          infile=trim(path_in)//'srb_rel3.0_shortwave_3hrly_'//cyear//cmon//'.binary'
          open (10,file=infile,access='direct',recl=nreg*2, form='unformatted',status='old')
          print*, trim(infile), ' input file is opened'
          read (10,rec=1) header, year, month, nvars, numgmt, ndays, ngrids,   &
              misval, vernum, vertab,&
              (Gmtime(hr),hr=1,8),(Jdays(day),day=1,Ndays),(Sclfac(I),I=1,11)
       else
          !open with smaller record length in order to read header information
          open (10,file=infile,access='direct',recl=4, form='unformatted',status='old')

          irec=62  !need to skip header var
          read (10,rec=irec) year
          call native_4byte_int(year,year)
          read (10,rec=irec+1) month
          call native_4byte_int(month,month)
          read (10,rec=irec+2) nvars
          call native_4byte_int(nvars,nvars)
          read (10,rec=irec+3) numgmt
          call native_4byte_int(numgmt,numgmt)
          read (10,rec=irec+4) ndays
          call native_4byte_int(ndays,ndays)
          read (10,rec=irec+5) ngrids
          call native_4byte_int(ngrids,ngrids)
          read (10,rec=irec+6) misval
          call native_4byte_real(misval,misval)
          read (10,rec=irec+7) vernum
          call native_4byte_real(vernum,vernum)
          read (10,rec=irec+8) vertab
          call native_4byte_real(vertab,vertab)
          irec=irec+8
          do i = 1,numgmt
             irec=irec+1
             read (10,rec=irec) gmtime(I)
             call native_4byte_int(gmtime(I),gmtime(I))
          end do
          do i = 1,ndays
             irec=irec+1
             read (10,rec=irec) Jdays(I)
             call native_4byte_int(Jdays(I),Jdays(I))
          end do
          do i = 1,nvars
             irec=irec+1
             read (10,rec=irec) Sclfac(I)
             call native_4byte_real(Sclfac(I),Sclfac(I))
          end do
          close (10) !close infile with smaller record length, reopen as usual
          open (10,file=infile,access='direct',recl=nreg*2, form='unformatted',status='old')
          print*, trim(infile), ' input file is opened'
          
       end if

!   Read all global variables
       do day=1,ndays
          do hr=1,8
             do iflx=1,11
                irec = 1+((day-1)*8*11)+(hr-1)*11+iFlx
                read (10,rec=irec) sw_val_in(:,hr,day,iflx)
!   Convert from big endian to little endian byte order, if necessary.
                if (little_endian) then
                  do reg=1,nreg
                     call native_2byte_int(sw_val_in(reg,hr,day,iflx),sw_val_in(reg,hr,day,iflx))
                  end do 
                end if
             end do 
          end do 
       end do 

!   Close input file
       close (10)

!   Convert desired variables to 1x1 degree global field
       do iflx=1,11

          if (var_list(iflx)) then

!   Open binary output file
             if (binary) then
                outfile=trim(path_out)//trim(var_header(iflx))//&
                    '3hrly_'//cyear//cmon//'.binary'
                open (15,file=outfile, form='unformatted', access='direct', &
                  recl=nlon*nlat*4, status='replace')
             end if
!   Open ascii output file
             if (ascii) then
                outfile2=trim(path_out)//trim(var_header(iflx))//&
                 '3hrly_'//cyear//cmon//'.ascii'
                open (25,file=outfile2, form='formatted', status='replace')
             end if
             jrec=0
             do day=1,ndays
                do hr=1,8

!   First keep missing values as missing values, otherwise get the correct
!   value for the variable by dividing by the scale factor.

                   sw_var=misval
                   do reg=1,nreg
                      if (sw_val_in(reg,hr,day,iflx)/= misval) then
                         sw_var(reg)=sw_val_in(reg,hr,day,iflx)/sclfac(iflx)
                      else
                         sw_var(reg)= misval
                      end if
                   end do

                   var0(:)=sw_var(:)
                   call regrid

!   Sample to screen:
                      if (day==14 .AND. hr==3) then
                     write (*,"(3A,I2.2,2x,A,I2.2)") 'Variable ', trim(var_header(iflx)), &
                            ' Hour = ', (hr-1)*3, 'Day = ', day
                      write (*,"(A,4x,5I10)") 'lon # = ', (lon,lon=100,104)
                      do lat=45,51
                         write (*,"(A,I3,5f10.3)") 'lat band # ',lat, &
                               (var2(lon,lat),lon=100,104)
                      end do
                      end if

!   Write output files, one field per file
!          binary output
                   jrec=jrec+1
                   if (binary) write (15,rec=jrec) var2

!          ascii output, starting at South Pole, Greenwich
                   if (ascii) then
                      do lat=1,nlat
                         write (25,"(15f10.3)") (var2(lon,lat),lon=1,nlon)
                      end do
                   end if
                end do
             end do
             if (binary) close (15)
             if (ascii) close (25)
             if (binary) print*, 'file ', trim(outfile), ' has been written'
             if (ascii) print*, 'file ', trim(outfile2), ' has been written'
          end if

       end do
       contains
!!!!****************************************************************!!!!
!
      subroutine regrid
!
!!!!****************************************************************!!!!
!
!   this routine is designed to regrid a field available on the nested
!   grid to an equal-angle 1x1 grid.  in this scheme, the larger
!   boxes (in longitude) are simply replicated into the 1x1 boxes
!   contained within them.
!
!!!!****************************************************************!!!!
!
      implicit none
!
      ereg=0
!
!   go through the boxes of nested grid, one at a time, and call
!   subroutine unnest to replicate the value of this box into an
!   appropriate number of 1x1 boxes (depending on the size of the
!   nested grid box).
!
      do nrg=1,nreg
!
        if (nrg.le.3) then
          call unnest(120)
        elseif (nrg.le.408) then
          call unnest(8)
        elseif (nrg.le.1308) then
          call unnest(4)
        elseif (nrg.le.5808) then
          call unnest(2)
        elseif (nrg.le.38208) then
          call unnest(1)
        elseif (nrg.le.42708) then
          call unnest(2)
        elseif (nrg.le.43608) then
          call unnest(4)
        elseif (nrg.le.44013) then
          call unnest(8)
        elseif (nrg.le.44016) then
          call unnest(120)
        end if
!
      end do  !  nrg
!
!   construct a 2-dimensional array from the 1-dimensional array
!   of the 1x1 field.
!
      do srg=1,sreg
!
        lat=(srg-1)/360+1
        lon=srg-(lat-1)*360
!
        var2(lon,lat)=var1(srg)
!
      end do  !  srg
!
      end subroutine regrid
!
!!!!****************************************************************!!!!
!
      subroutine unnest(mreg)
!
!!!!****************************************************************!!!!
!
      implicit none
!
      integer, intent(in) :: mreg
!
      breg=ereg+1
      ereg=breg+mreg-1
!
!   replicate the value from one box of the nested grid into all
!   1x1 boxes contained in it.
!
      do srg=breg,ereg
        var1(srg)=var0(nrg)
      end do  !  nrg
!
      end subroutine unnest
!
!!!!****************************************************************!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!           FILE: SUBR_native_4byte_real.f90
!     SUBPROGRAM: native_4byte_real
!
!         AUTHOR: David Stepaniak, NCAR/CGD/CAS
! DATE INITIATED: 29 April 2003
!  LAST MODIFIED: 29 April 2003
!
!       SYNOPSIS: Converts a 32 bit, 4 byte, REAL from big Endian to
!                 little Endian, or conversely from little Endian to big
!                 Endian.
!
!    DESCRIPTION: This subprogram allows one to convert a 32 bit, 4 byte,
!                 REAL data element that was generated with, say, a big
!                 Endian processor (e.g. Sun/sparc, SGI/R10000, etc.) to its
!                 equivalent little Endian representation for use on little
!                 Endian processors (e.g. PC/Pentium running Linux). The
!                 converse, little Endian to big Endian, also holds.
!                 This conversion is accomplished by writing the 32 bits of
!                 the REAL data element into a generic 32 bit INTEGER space
!                 with the TRANSFER intrinsic, reordering the 4 bytes with
!                 the MVBITS intrinsic, and writing the reordered bytes into
!                 a new 32 bit REAL data element, again with the TRANSFER
!                 intrinsic. The following schematic illustrates the
!                 reordering process
!
!
!                  --------    --------    --------    --------
!                 |    D   |  |    C   |  |    B   |  |    A   |  4 Bytes
!                  --------    --------    --------    --------
!                                                             |
!                                                              -> 1 bit
!                                       ||
!                                     MVBITS
!                                       ||
!                                       \/
!
!                  --------    --------    --------    --------
!                 |    A   |  |    B   |  |    C   |  |    D   |  4 Bytes
!                  --------    --------    --------    --------
!                         |           |           |           |
!                         24          16          8           0   <- bit
!                                                                 position
!
!          INPUT: realIn,  a single 32 bit, 4 byte REAL data element.
!         OUTPUT: realOut, a single 32 bit, 4 byte REAL data element, with
!                 reverse byte order to that of realIn.
!    RESTRICTION: It is assumed that the default REAL data element is
!                 32 bits / 4 bytes.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  SUBROUTINE native_4byte_real( realIn, realOut )
  IMPLICIT NONE
  REAL, INTENT(IN)                              :: realIn
                                                   ! a single 32 bit, 4 byte
                                                   ! REAL data element
  REAL, INTENT(OUT)                             :: realOut
                                                   ! a single 32 bit, 4 byte
                                                   ! REAL data element, with
                                                   ! reverse byte order to
                                                   ! that of realIn
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Local variables (generic 32 bit INTEGER spaces):

  INTEGER                                       :: i_element
  INTEGER                                       :: i_element_br
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Transfer 32 bit of realIn to generic 32 bit INTEGER space:
  i_element = TRANSFER( realIn, 0 )
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Reverse order of 4 bytes in 32 bit INTEGER space:
  CALL MVBITS( i_element, 24, 8, i_element_br, 0  )
  CALL MVBITS( i_element, 16, 8, i_element_br, 8  )
  CALL MVBITS( i_element,  8, 8, i_element_br, 16 )
  CALL MVBITS( i_element,  0, 8, i_element_br, 24 )
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Transfer reversed order bytes to 32 bit REAL space (realOut):
  realOut = TRANSFER( i_element_br, 0.0 )
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  END SUBROUTINE
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  SUBROUTINE native_4byte_int( integerIn, integerOut )
  
! This subroutine is based on native_4byte_real, but provides the 
! conversions for integer variables.

  IMPLICIT NONE
  integer, INTENT(IN)                              :: integerIn
                                                   ! a single 32 bit, 4 byte
  integer, INTENT(OUT)                             :: integerOut
                                                   ! a single 32 bit, 4 byte
                                                   ! reverse byte order to
                                                   ! that of integerIn
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Local variables (generic 32 bit INTEGER spaces):

  INTEGER                                       :: i_element
  INTEGER                                       :: i_element_br

  i_element=integerIn
! Reverse order of 4 bytes in 32 bit INTEGER space:
  CALL MVBITS( i_element, 24, 8, i_element_br, 0  )
  CALL MVBITS( i_element, 16, 8, i_element_br, 8  )
  CALL MVBITS( i_element,  8, 8, i_element_br, 16 )
  CALL MVBITS( i_element,  0, 8, i_element_br, 24 )
  integerOut=i_element_br
  END SUBROUTINE
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  SUBROUTINE native_2byte_int( integerIn, integerOut )

! This subroutine is based on native_4byte_real, but provides the
! conversions for integer variables.

  IMPLICIT NONE
  integer (kind=2), INTENT(IN)  :: integerIn
                                                   ! a single 16 bit, 2 byte
  integer (kind=2), INTENT(OUT) :: integerOut
                                                   ! a single 16 bit, 2 byte
                                                   ! reverse byte order to
                                                   ! that of integerIn
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Local variables (generic 16 bit INTEGER spaces):

  INTEGER (kind=2) :: i_element
  INTEGER (kind=2) :: i_element_br

  i_element=integerIn

  CALL MVBITS( i_element,8, 8, i_element_br, 0  )
  CALL MVBITS( i_element,0, 8, i_element_br, 8  )

  integerOut=i_element_br

  End SUBROUTINE native_2byte_int 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      end program read_srb_rel3_sw_3hrly
